#include "precompiled.h"

#include "gdal_priv.h"
#include "ogr_spatialref.h"

std::string MakeKey(const char* name, const std::string& value)
{
	std::string key(name);
	key += '=';
	key += value;
	return key;
}


static void EncodeLatLon(double lat, double lon, char* BLOCKA_loc)
{
	sprintf(BLOCKA_loc, "%+#010.6f%+#011.6f", lat, lon);
	assert(strlen(BLOCKA_loc) == 21);
}

void ToOptions(std::vector<std::string>& options)
{
	// (set elsewhere)
	size_t m_height;
	double m_ul_lat, m_ul_lon, m_lr_lat, m_lr_lon;

	options.push_back(MakeKey("ICORDS", "G"));	// Image Coordinate Representation
	// don't pass IGEOLO directly; we'll set it via GDAL
	// note: NSIF doesn't seem to be able to store the UTM zone. GDAL writes
	// it into an .aux.xml, but that's not transferred to the xxxxxxx. we therefore
	// need to use lat/lon (degree) coordinates.


	// BLOCKA TRE
	{
		options.push_back(MakeKey("BLOCKA_BLOCK_COUNT", "1"));
		options.push_back(MakeKey("BLOCKA_BLOCK_INSTANCE_01", "01"));
		options.push_back(MakeKey("BLOCKA_N_GRAY_01", "00000"));
		char l_lines[6];
		sprintf(l_lines, "%05d", m_height);
		options.push_back(MakeKey("BLOCKA_L_LINES_01", l_lines));
		options.push_back(MakeKey("BLOCKA_LAYOVER_ANGLE_01", "   "));
		options.push_back(MakeKey("BLOCKA_SHADOW_ANGLE_01", "   "));
		char loc[22];
		EncodeLatLon(m_lr_lat, m_ul_lon, loc);
		options.push_back(MakeKey("BLOCKA_FRLC_LOC_01", loc));
		EncodeLatLon(m_lr_lat, m_lr_lon, loc);
		options.push_back(MakeKey("BLOCKA_LRLC_LOC_01", loc));
		EncodeLatLon(m_ul_lat, m_lr_lon, loc);
		options.push_back(MakeKey("BLOCKA_LRFC_LOC_01", loc));
		EncodeLatLon(m_ul_lat, m_ul_lon, loc);
		options.push_back(MakeKey("BLOCKA_FRFC_LOC_01", loc));
	}
}



struct GeoInfo
{
	void GetCorners(double& ul_lat, double& ul_lon, double& lr_lat, double& lr_lon) const;

	class Impl;
	shared_ptr<Impl> impl;
};


class GeoInfo::Impl
{
public:
	GeoInfo::Impl(GDALDataset* dataset)
		: m_width(dataset->GetRasterXSize()), m_height(dataset->GetRasterYSize())
	{
		const char* proj = dataset->GetProjectionRef();
		OGRSpatialReference sr(proj);

		double adfTransform[6];
		dataset->GetGeoTransform(adfTransform);

		// (it seems strange to clone a spatial reference and then create a
		// transform between that and the original, but that's what
		// gdalinfo does. we check for coordTransform == 0 below.)
		OGRSpatialReference* sr_latlong = sr.CloneGeogCS();
		OGRCoordinateTransformation* coordTransform = OGRCreateCoordinateTransformation(&sr, sr_latlong);
		delete sr_latlong;

		PixelToLatLong(0, 0, adfTransform, coordTransform, m_ul_lat, m_ul_lon);
		PixelToLatLong(m_width, m_height, adfTransform, coordTransform, m_lr_lat, m_lr_lon);

		delete coordTransform;
	}

	void Apply(GDALDataset* dataset)
	{
		dataset->SetProjection("WGS84");

		// from GDAL documentation:
		// 0: top left x
		// 1: w-e pixel resolution
		// 2: rotation, 0 if image is "north up"
		// 3: top left y
		// 4: rotation, 0 if image is "north up"
		// 5: n-s pixel resolution
		double adfTransform[6];
		adfTransform[0] = m_ul_lat;
		adfTransform[1] = fabs(m_ul_lat - m_lr_lat) / m_width;
		adfTransform[2] = 0.0;
		adfTransform[3] = m_ul_lon;
		adfTransform[4] = fabs(m_ul_lon - m_lr_lon) / m_height;
		adfTransform[5] = 0.0;
		// when creating the dataset, we pass options that cause a complete
		// BLOCKA TRE to be written. the final part of GDAL Create calls
		// GDALOpen, which ends up invoking NITFReadBLOCKA_GCPs. the "GCPs"
		// are actually image corners and are read from BLOCKA, after which
		// the image ICORDS value is changed from G to D. if we now call
		// SetGeoTransform then NITFWriteIGEOLO will fail because it
		// currently requires ICORDS to be one of {G,N,S}. since BLOCKA
		// completely specifies the coordinates (at higher precision than
		// IGEOLO), we will skip SetGeoTransform until this undesirable
		// behavior in GDAL (currently 1.5.2) is fixed.
//		dataset->SetGeoTransform(adfTransform);
	}

	// get latitude/longitude of upper left and lower right corners
	// (coordinate format: see PixelToLatLong)
	void GetCorners(double& ul_lat, double& ul_lon, double& lr_lat, double& lr_lon)
	{
		ul_lat = m_ul_lat;
		ul_lon = m_ul_lon;
		lr_lat = m_lr_lat;
		lr_lon = m_lr_lon;
	}

private:
	// @return latitude/longitude in degrees
	static void PixelToLatLong(size_t x, size_t y, double* adfTransform, OGRCoordinateTransformation* coordTransform, double& lat, double& lon)
	{
		double geo_x = adfTransform[0] + adfTransform[1]*x + adfTransform[2]*y;
		double geo_y = adfTransform[3] + adfTransform[4]*x + adfTransform[5]*y;
		if(coordTransform)
			coordTransform->Transform(1, &geo_x, &geo_y, 0);

		assert(-180.0 <= geo_x && geo_x <= 180.0);
		assert(-90.0 <= geo_y && geo_y <= 90.0);
		lat = geo_x;
		lon = geo_y;
	}

	size_t m_width;
	size_t m_height;
	double m_ul_lat, m_ul_lon, m_lr_lat, m_lr_lon;
};


void GeoInfo::GetCorners(double& ul_lat, double& ul_lon, double& lr_lat, double& lr_lon) const
{
	impl->GetCorners(ul_lat, ul_lon, lr_lat, lr_lon);
}



static GDALDataType TypeFromComponentSize(size_t componentSize)
{
	switch(componentSize)
	{
	case 1:
		return GDT_Byte;
	case 2:
		return GDT_Int16;
	case 4:
		return GDT_Float32;
	case 8:
		return GDT_Float64;
	default:
		throw std::runtime_error("unsupported componentSize");
	}
}

void WriteNITF(size_t width, size_t height, size_t numComponents, size_t componentSize, void* pixels, const char* pathname, const std::vector<std::string>* options, const GeoInfo& geoInfo)
{
	assert(numComponents == 1);	// not yet correctly supported

	const int w = (int)width, h = (int)height, c = (int)numComponents;
	const GDALDataType dataType = TypeFromComponentSize(componentSize);

	const char* driverName = "NITF";
	GDALDriver* driver = (GDALDriver*)GDALGetDriverByName(driverName);
	if(!driver)
		throw std::runtime_error("unknown driverName");

	char** options2 = 0;
	if(options)
	{
		for(size_t i = 0; i < options->size(); i++)
			options2 = CSLAddString(options2, (*options)[i].c_str());
	}

	GDALDataset* dataset = driver->Create(pathname, w, h, c, dataType, options2);
	if(!dataset)
		throw std::runtime_error("dataset creation failed");

	geoInfo.impl->Apply(dataset);

	size_t pos = 0;
	for(size_t i = 1; i <= numComponents; i++)
	{
		GDALRasterBand* band = dataset->GetRasterBand((int)i);
		band->RasterIO(GF_Write, 0, 0, w, h, (u8*)pixels+pos, w, h, dataType, 0, 0);

		pos += width*height;
	}

	delete dataset;
	dataset = 0;
}
